'use strict';

var fetch$1 = require('node-fetch');
var sodium = require('libsodium-wrappers');
var nodeTfhe = require('node-tfhe');
var bigintBuffer = require('bigint-buffer');
var web3Validator = require('web3-validator');
var createKeccakHash = require('keccak');
var nodeKms = require('node-kms');

/******************************************************************************
Copyright (c) Microsoft Corporation.

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES WITH
REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY SPECIAL, DIRECT,
INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.
***************************************************************************** */
/* global Reflect, Promise */


function __awaiter(thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
}

function __generator(thisArg, body) {
    var _ = { label: 0, sent: function() { if (t[0] & 1) throw t[1]; return t[1]; }, trys: [], ops: [] }, f, y, t, g;
    return g = { next: verb(0), "throw": verb(1), "return": verb(2) }, typeof Symbol === "function" && (g[Symbol.iterator] = function() { return this; }), g;
    function verb(n) { return function (v) { return step([n, v]); }; }
    function step(op) {
        if (f) throw new TypeError("Generator is already executing.");
        while (g && (g = 0, op[0] && (_ = 0)), _) try {
            if (f = 1, y && (t = op[0] & 2 ? y["return"] : op[0] ? y["throw"] || ((t = y["return"]) && t.call(y), 0) : y.next) && !(t = t.call(y, op[1])).done) return t;
            if (y = 0, t) op = [op[0] & 2, t.value];
            switch (op[0]) {
                case 0: case 1: t = op; break;
                case 4: _.label++; return { value: op[1], done: false };
                case 5: _.label++; y = op[1]; op = [0]; continue;
                case 7: op = _.ops.pop(); _.trys.pop(); continue;
                default:
                    if (!(t = _.trys, t = t.length > 0 && t[t.length - 1]) && (op[0] === 6 || op[0] === 2)) { _ = 0; continue; }
                    if (op[0] === 3 && (!t || (op[1] > t[0] && op[1] < t[3]))) { _.label = op[1]; break; }
                    if (op[0] === 6 && _.label < t[1]) { _.label = t[1]; t = op; break; }
                    if (t && _.label < t[2]) { _.label = t[2]; _.ops.push(op); break; }
                    if (t[2]) _.ops.pop();
                    _.trys.pop(); continue;
            }
            op = body.call(thisArg, _);
        } catch (e) { op = [6, e]; y = 0; } finally { f = t = 0; }
        if (op[0] & 5) throw op[1]; return { value: op[0] ? op[1] : void 0, done: true };
    }
}

var fromHexString = function (hexString) {
    var arr = hexString.replace(/^(0x)/, '').match(/.{1,2}/g);
    if (!arr)
        return new Uint8Array();
    return Uint8Array.from(arr.map(function (byte) { return parseInt(byte, 16); }));
};
var toHexString = function (bytes) {
    return bytes.reduce(function (str, byte) { return str + byte.toString(16).padStart(2, '0'); }, '');
};
var bytesToBigInt = function (byteArray) {
    if (!byteArray || (byteArray === null || byteArray === void 0 ? void 0 : byteArray.length) === 0) {
        return BigInt(0);
    }
    var buffer = Buffer.from(byteArray);
    var result = bigintBuffer.toBigIntBE(buffer);
    return result;
};
var clientKeyDecryptor = function (clientKeySer) {
    var clientKey = nodeTfhe.TfheClientKey.deserialize(clientKeySer);
    return {
        decryptBool: function (ciphertext) {
            return nodeTfhe.FheBool.deserialize(fromHexString(ciphertext)).decrypt(clientKey);
        },
        decrypt4: function (ciphertext) {
            return nodeTfhe.FheUint4.deserialize(fromHexString(ciphertext)).decrypt(clientKey);
        },
        decrypt8: function (ciphertext) {
            return nodeTfhe.FheUint8.deserialize(fromHexString(ciphertext)).decrypt(clientKey);
        },
        decrypt16: function (ciphertext) {
            return nodeTfhe.FheUint16.deserialize(fromHexString(ciphertext)).decrypt(clientKey);
        },
        decrypt32: function (ciphertext) {
            return nodeTfhe.FheUint32.deserialize(fromHexString(ciphertext)).decrypt(clientKey);
        },
        decrypt64: function (ciphertext) {
            return nodeTfhe.FheUint64.deserialize(fromHexString(ciphertext)).decrypt(clientKey);
        },
        decryptAddress: function (ciphertext) {
            var hex = nodeTfhe.FheUint160.deserialize(fromHexString(ciphertext))
                .decrypt(clientKey)
                .toString(16);
            while (hex.length < 40) {
                hex = '0' + hex;
            }
            return '0x' + hex;
        },
    };
};

function decodeAbiBytes(hex) {
    if (hex.startsWith('0x')) {
        hex = hex.substring(2);
    }
    // The data offset is the first 32 bytes, which points to the start of the data (skip this part)
    var dataOffset = parseInt(hex.substring(0, 64), 16) * 2; // Convert offset from bytes to characters
    // The length of the data is the next 32 bytes, starting from the offset
    var dataLength = parseInt(hex.substring(dataOffset, dataOffset + 64), 16) * 2; // Convert length from bytes to characters
    // Extract the data
    var data = hex.substring(dataOffset + 64, dataOffset + 64 + dataLength);
    // Convert the extracted data from hex to Uint8Array
    var bytes = new Uint8Array(data.length / 2);
    for (var i = 0, j = 0; i < data.length; i += 2, j++) {
        bytes[j] = parseInt(data.substring(i, i + 2), 16);
    }
    return bytes;
}
var fetchJSONRPC = function (url, options) { return __awaiter(void 0, void 0, void 0, function () {
    var response, data, hexResult, decodedBytes, error_1;
    return __generator(this, function (_a) {
        switch (_a.label) {
            case 0:
                _a.trys.push([0, 3, , 4]);
                return [4 /*yield*/, fetch(url, options)];
            case 1:
                response = _a.sent();
                if (!response.ok) {
                    throw new Error("HTTP error! status: ".concat(response.status));
                }
                return [4 /*yield*/, response.json()];
            case 2:
                data = _a.sent();
                if (data && data.result) {
                    hexResult = data.result;
                    decodedBytes = decodeAbiBytes(hexResult);
                    return [2 /*return*/, "0x".concat(toHexString(decodedBytes))];
                }
                else {
                    console.error('No result from blockchain call');
                }
                return [3 /*break*/, 4];
            case 3:
                error_1 = _a.sent();
                console.error('Error performing eth_call:', error_1);
                return [3 /*break*/, 4];
            case 4: return [2 /*return*/];
        }
    });
}); };

var getPublicKeyCallParams = function () { return ({
    to: '0x000000000000000000000000000000000000005d',
    data: '0xd9d47bb001',
}); };
var getChainIdFromNetwork = function (url) { return __awaiter(void 0, void 0, void 0, function () {
    var payload, options;
    return __generator(this, function (_a) {
        payload = {
            jsonrpc: '2.0',
            method: 'eth_chainId',
            params: [],
            id: 1,
        };
        options = {
            method: 'POST',
            headers: {
                'Content-Type': 'application/json',
            },
            body: JSON.stringify(payload),
        };
        return [2 /*return*/, Number(fetchJSONRPC(url, options))];
    });
}); };
// Define the function to perform the eth_call
var getPublicKeyFromNetwork = function (url) { return __awaiter(void 0, void 0, void 0, function () {
    var payload, options;
    return __generator(this, function (_a) {
        payload = {
            jsonrpc: '2.0',
            method: 'eth_call',
            params: [getPublicKeyCallParams(), 'latest'],
            id: 1,
        };
        options = {
            method: 'POST',
            headers: {
                'Content-Type': 'application/json',
            },
            body: JSON.stringify(payload),
        };
        return [2 /*return*/, fetchJSONRPC(url, options)];
    });
}); };
// Define the function to perform the eth_call
var getPublicKeyFromCoprocessor = function (url) { return __awaiter(void 0, void 0, void 0, function () {
    var payload, options;
    return __generator(this, function (_a) {
        payload = {
            jsonrpc: '2.0',
            method: 'eth_getPublicFhevmKey',
            params: [],
            id: 1,
        };
        options = {
            method: 'POST',
            headers: {
                'Content-Type': 'application/json',
            },
            body: JSON.stringify(payload),
        };
        return [2 /*return*/, fetchJSONRPC(url, options)];
    });
}); };

var ENCRYPTION_TYPES = {
    1: 0,
    4: 1,
    8: 2,
    16: 3,
    32: 4,
    64: 5,
    128: 6,
    160: 7,
    256: 8,
    512: 9,
    1024: 10,
    2048: 11,
};

var checkEncryptedValue = function (value, bits) {
    if (value == null)
        throw new Error('Missing value');
    var limit;
    if (bits >= 8) {
        limit = BigInt("0x".concat(new Array(bits / 8).fill(null).reduce(function (v) { return "".concat(v, "ff"); }, '')));
    }
    else {
        limit = BigInt(Math.pow(2, bits) - 1);
    }
    if (typeof value !== 'number' && typeof value !== 'bigint')
        throw new Error('Value must be a number or a bigint.');
    if (value > limit) {
        throw new Error("The value exceeds the limit for ".concat(bits, "bits integer (").concat(limit.toString(), ")."));
    }
};
var createEncryptedInput = function (tfheCompactPublicKey, coprocessorUrl) {
    return function (contractAddress, callerAddress) {
        if (!tfheCompactPublicKey)
            throw new Error('Your instance has been created without the public blockchain key.');
        if (!web3Validator.isAddress(contractAddress)) {
            throw new Error('Contract address is not a valid address.');
        }
        if (!web3Validator.isAddress(callerAddress)) {
            throw new Error('User address is not a valid address.');
        }
        var publicKey = tfheCompactPublicKey;
        var values = [];
        var bits = [];
        return {
            addBool: function (value) {
                if (value == null)
                    throw new Error('Missing value');
                if (typeof value !== 'boolean' &&
                    typeof value !== 'number' &&
                    typeof value !== 'bigint')
                    throw new Error('The value must be a boolean, a number or a bigint.');
                if ((typeof value !== 'bigint' || typeof value !== 'number') &&
                    Number(value) > 1)
                    throw new Error('The value must be 1 or 0.');
                values.push(BigInt(value));
                bits.push(1);
                return this;
            },
            add4: function (value) {
                checkEncryptedValue(value, 4);
                values.push(BigInt(value));
                bits.push(4);
                return this;
            },
            add8: function (value) {
                checkEncryptedValue(value, 8);
                values.push(BigInt(value));
                bits.push(8);
                return this;
            },
            add16: function (value) {
                checkEncryptedValue(value, 16);
                values.push(BigInt(value));
                bits.push(16);
                return this;
            },
            add32: function (value) {
                checkEncryptedValue(value, 32);
                values.push(BigInt(value));
                bits.push(32);
                return this;
            },
            add64: function (value) {
                checkEncryptedValue(value, 64);
                values.push(BigInt(value));
                bits.push(64);
                return this;
            },
            add128: function (value) {
                checkEncryptedValue(value, 128);
                values.push(BigInt(value));
                bits.push(128);
                return this;
            },
            addAddress: function (value) {
                if (!web3Validator.isAddress(value)) {
                    throw new Error('The value must be a valid address.');
                }
                values.push(BigInt(value));
                bits.push(160);
                return this;
            },
            addBytes256: function (value) {
                var bigIntValue = bytesToBigInt(value);
                checkEncryptedValue(bigIntValue, 2048);
                values.push(bigIntValue);
                bits.push(2048);
                return this;
            },
            getValues: function () {
                return values;
            },
            getBits: function () {
                return bits;
            },
            resetValues: function () {
                values.length = 0;
                bits.length = 0;
                return this;
            },
            encrypt: function () {
                if (bits.reduce(function (total, v) { return total + v; }, 0) > 2048) {
                    throw new Error('Too many bits in provided values. Maximum is 2048.');
                }
                var encrypted;
                if (bits.some(function (v) { return v === 2048; })) {
                    encrypted = nodeTfhe.CompactFheUint2048List.encrypt_with_compact_public_key(values, publicKey);
                }
                else {
                    encrypted = nodeTfhe.CompactFheUint160List.encrypt_with_compact_public_key(values, publicKey);
                }
                var data = encrypted.serialize();
                var hash = createKeccakHash('keccak256')
                    .update(Buffer.from(data))
                    .digest();
                // const encrypted = ProvenCompactFheUint160List.encrypt_with_compact_public_key(
                //   values,
                //   publicZkParams,
                //   publicKey,
                //   ZkComputeLoad.Proof,
                // );
                var inputs = bits.map(function (v, i) {
                    var dataWithIndex = new Uint8Array(hash.length + 1);
                    dataWithIndex.set(hash, 0);
                    dataWithIndex.set([i], hash.length);
                    var finalHash = createKeccakHash('keccak256')
                        .update(Buffer.from(dataWithIndex))
                        .digest();
                    var dataInput = new Uint8Array(32);
                    dataInput.set(finalHash, 0);
                    dataInput.set([i, ENCRYPTION_TYPES[v], 0], 29);
                    return dataInput;
                });
                return {
                    inputs: inputs,
                    data: data,
                };
            },
            send: function () {
                return __awaiter(this, void 0, void 0, function () {
                    var encrypted, ciphertext, data, payload, options, response;
                    return __generator(this, function (_a) {
                        switch (_a.label) {
                            case 0:
                                if (!coprocessorUrl)
                                    throw new Error('Coprocessor URL not provided');
                                encrypted = nodeTfhe.CompactFheUint160List.encrypt_with_compact_public_key(values, publicKey);
                                ciphertext = encrypted.serialize();
                                data = new Uint8Array(1 + bits.length + ciphertext.length);
                                data.set([data.length], 0);
                                data.set(bits, 1);
                                data.set(ciphertext, data.length + 1);
                                payload = {
                                    jsonrpc: '2.0',
                                    method: 'eth_addUserCiphertext',
                                    params: [toHexString(data), contractAddress, callerAddress],
                                    id: 1,
                                };
                                options = {
                                    method: 'POST',
                                    headers: {
                                        'Content-Type': 'application/json',
                                    },
                                    body: JSON.stringify(payload),
                                };
                                return [4 /*yield*/, fetchJSONRPC(coprocessorUrl, options)];
                            case 1:
                                response = _a.sent();
                                if (!response)
                                    throw new Error('Invalid input');
                                return [2 /*return*/, JSON.parse(response)];
                        }
                    });
                });
            },
        };
    };
};

var createEIP712 = function (chainId) {
    return function (publicKey, verifyingContract, delegatedAccount) {
        if (!web3Validator.isAddress(verifyingContract))
            throw new Error('Invalid contract address.');
        if (delegatedAccount && !web3Validator.isAddress(delegatedAccount))
            throw new Error('Invalid delegated account.');
        var msgParams = {
            types: {
                // This refers to the domain the contract is hosted on.
                EIP712Domain: [
                    { name: 'name', type: 'string' },
                    { name: 'version', type: 'string' },
                    { name: 'chainId', type: 'uint256' },
                    { name: 'verifyingContract', type: 'address' },
                ],
                // Refer to primaryType.
                Reencrypt: [{ name: 'publicKey', type: 'bytes32' }],
            },
            // This defines the message you're proposing the user to sign, is dapp-specific, and contains
            // anything you want. There are no required fields. Be as explicit as possible when building out
            // the message schema.
            // This refers to the keys of the following types object.
            primaryType: 'Reencrypt',
            domain: {
                // Give a user-friendly name to the specific contract you're signing for.
                name: 'Authorization token',
                // This identifies the latest version.
                version: '1',
                // This defines the network, in this case, Mainnet.
                chainId: chainId,
                // // Add a verifying contract to make sure you're establishing contracts with the proper entity.
                verifyingContract: verifyingContract,
            },
            message: {
                publicKey: "0x".concat(publicKey),
            },
        };
        if (delegatedAccount) {
            msgParams.message.delegatedAccount = delegatedAccount;
            msgParams.types.Reencrypt.push({
                name: 'delegatedAccount',
                type: 'address',
            });
        }
        return msgParams;
    };
};
var generateKeypair = function () {
    var keypair = nodeKms.cryptobox_keygen();
    return {
        publicKey: toHexString(nodeKms.cryptobox_pk_to_u8vec(nodeKms.cryptobox_get_pk(keypair))),
        privateKey: toHexString(nodeKms.cryptobox_sk_to_u8vec(keypair)),
    };
};

var reencryptRequest = function (reencryptionUrl) {
    return function (handle, privateKey, publicKey, signature, contractAddress, userAddress) { return __awaiter(void 0, void 0, void 0, function () {
        return __generator(this, function (_a) {
            if (!reencryptionUrl)
                throw new Error('You must provide a reencryption URL.');
            // const response = await fetch(`${reencryptUrl}`);
            return [2 /*return*/, BigInt(10)];
        });
    }); };
};

var createInstance = function (config) { return __awaiter(void 0, void 0, void 0, function () {
    var networkUrl, reencryptionUrl, coprocessorUrl, chainId, publicKey, tfheCompactPublicKey, buff;
    return __generator(this, function (_a) {
        switch (_a.label) {
            case 0: return [4 /*yield*/, sodium.ready];
            case 1:
                _a.sent();
                networkUrl = config.networkUrl, reencryptionUrl = config.reencryptionUrl, coprocessorUrl = config.coprocessorUrl;
                chainId = config.chainId;
                publicKey = config.publicKey;
                if (typeof chainId !== 'number')
                    throw new Error('chainId must be a number.');
                if (!(coprocessorUrl && !publicKey)) return [3 /*break*/, 3];
                return [4 /*yield*/, getPublicKeyFromCoprocessor(coprocessorUrl)];
            case 2:
                publicKey = _a.sent();
                return [3 /*break*/, 5];
            case 3:
                if (!(networkUrl && !publicKey)) return [3 /*break*/, 5];
                return [4 /*yield*/, getPublicKeyFromNetwork(networkUrl)];
            case 4:
                publicKey = _a.sent();
                _a.label = 5;
            case 5:
                if (!(networkUrl && !chainId)) return [3 /*break*/, 7];
                return [4 /*yield*/, getChainIdFromNetwork(networkUrl)];
            case 6:
                chainId = _a.sent();
                _a.label = 7;
            case 7:
                if (publicKey && typeof publicKey !== 'string')
                    throw new Error('publicKey must be a string');
                if (publicKey) {
                    buff = fromHexString(publicKey);
                    try {
                        tfheCompactPublicKey = nodeTfhe.TfheCompactPublicKey.deserialize(buff);
                    }
                    catch (e) {
                        throw new Error('Invalid public key (deserialization failed)');
                    }
                }
                return [2 /*return*/, {
                        createEncryptedInput: createEncryptedInput(tfheCompactPublicKey, coprocessorUrl),
                        generateKeypair: generateKeypair,
                        createEIP712: createEIP712(chainId),
                        reencrypt: reencryptRequest(reencryptionUrl),
                    }];
        }
    });
}); };

var createTfheKeypair = function () {
    var block_params = new nodeTfhe.ShortintParameters(nodeTfhe.ShortintParametersName.PARAM_MESSAGE_2_CARRY_2_COMPACT_PK_PBS_KS);
    // const configBuilder = new TfheConfigBuilder();
    // const config = configBuilder.use_custom_parameters(block_params).build();
    var config = nodeTfhe.TfheConfigBuilder.default()
        .use_custom_parameters(block_params)
        .build();
    var clientKey = nodeTfhe.TfheClientKey.generate(config);
    var publicKey = nodeTfhe.TfheCompactPublicKey.new(clientKey);
    publicKey = nodeTfhe.TfheCompactPublicKey.deserialize(publicKey.serialize());
    return { clientKey: clientKey, publicKey: publicKey };
};
var createTfhePublicKey = function () {
    var publicKey = createTfheKeypair().publicKey;
    return toHexString(publicKey.serialize());
};

//@ts-check
if (!global.fetch) {
    global.fetch = fetch$1;
    globalThis.Headers = fetch$1.Headers;
    globalThis.Request = fetch$1.Request;
    globalThis.Response = fetch$1.Response;
}

exports.clientKeyDecryptor = clientKeyDecryptor;
exports.createInstance = createInstance;
exports.createTfheKeypair = createTfheKeypair;
exports.createTfhePublicKey = createTfhePublicKey;
exports.getPublicKeyCallParams = getPublicKeyCallParams;
